"""Root directory for the scm package"""
# scm/__init__.py

__app_name__ = "scm"
__version__ = "0.2.4"

(
 SUCCESS, 
 DIR_ERROR, 
 FILE_ERROR, 
 IO_ERROR, 
 VALIDATION_ERROR, 
 OS_ERROR    
) = range(6)

ERRORS = {
    DIR_ERROR: "Config directory error", 
    FILE_ERROR: "Config file error", 
    IO_ERROR: "Error during the IO Operation", 
    VALIDATION_ERROR: "Error during the validation",
    OS_ERROR: "Error from the OS system module"
    
}


"""Module provides the CLI package"""
# scm/cli.py
from multiprocessing.dummy import current_process
import os
import logging
from scm_config.defaults import SETTINGS_DATA
from typing import Optional
from ordered_set import OrderedSet
import typer
import hashlib
import functools
from scm_config import __version__, __app_name__, default_config
from scm_config.default_config import read_json, create_def_directory, create_def_files, get_diff_hash_res, check_if_receipe_exists, get_user_settings, get_user_defined_resources, validate_unsupported_resources, gen_command, run_os_command, _hash_fun,_get_diff_hash, _write_json
from dynaconf import Dynaconf
from dynaconf.utils.boxing import DynaBox
from dynaconf.vendor.toml.decoder import TomlDecodeError
from ordered_set import OrderedSet
from collections import OrderedDict
from scm_config import defaults
import json

# BASIC LOGGING MESSAGE
logging.basicConfig(format='[%(levelname)s][%(asctime)s]::%(message)s',
                    datefmt="%m-%d-%Y %H:%M:%S", level=logging.INFO)
logger = logging.getLogger()

# Global variables
def_settings_dict = {}

app = typer.Typer()


def _version(value: bool) -> None:
    if value:
        typer.echo(f"{__app_name__} - v{__version__}")
        raise typer.Exit()


@app.command()
def init(
    file: str = typer.Option(
        str(default_config.CONFIG_FILE_PATH),
        "--default-file",
        "-f"
    )
) -> None:
    global def_settings_dict
     
    with open(file, "w") as outfile:
        json.dump(SETTINGS_DATA, outfile)
    
    
    def_settings_dict = read_json(file)['default']

    os.environ['ROOT_PATH_FOR_DYNACONF'] = def_settings_dict.get(
        'CONFIG_DIR', os.path.join(os.getcwd(), "config"))

    is_dir_created = create_def_directory(def_settings_dict)

    if not is_dir_created:
        logging.warning("default directories creation failed")
        raise typer.Exit()

    is_files_created = create_def_files(def_settings_dict)

    if not is_files_created:
        logging.warning("default files creation failed")
        raise typer.Exit()

    
    return 0

# create command


@app.command()
def create(
    receipe: str = typer.Option(...),
    force: bool = typer.Option(False)
) -> None:

    init(default_config.CONFIG_FILE_PATH)
    con_dir, def_file = def_settings_dict.get('CONFIG_DIR'), f"{receipe}.toml"
    con_dir_full = os.path.join(os.getcwd(), con_dir)
    def_file_full = os.path.join(con_dir_full, def_file)
    
    

    if not check_if_receipe_exists(receipe) or force:
        with open(def_file_full, mode="w") as f:
            f.write(f"# Place holder for creating the scm {receipe} receipe")
    else:
        logging.warning(
            f"`{receipe}` configuration file exists, use --force to override.")
        raise typer.Exit()

    
    
    return 0


@app.command()
def info(
    receipe: str = typer.Option(...)
) -> None:

    init(default_config.CONFIG_FILE_PATH)
    con_dir, def_file = def_settings_dict.get('CONFIG_DIR'), f"{receipe}.toml"
    con_dir_full = os.path.join(os.getcwd(), con_dir)
    def_file_full = os.path.join(con_dir_full, def_file)

    if not check_if_receipe_exists(receipe):
        logging.warning(
            f"receipe {receipe} doesn't exist in the config directory")
        logging.warning("use `scm create` for the receipe creation")
        raise typer.Exit()

    user_settings = get_user_settings(receipe)
    user_resources = get_user_defined_resources(user_settings)

    for res in user_resources:
        if res in user_settings:
            for i in user_settings[res]:
                logging.info(f"{res}.{i} - {user_settings[res][i]}")

    return 0

# validate command


@app.command()
def validate(
    receipe: str = typer.Option(...)
) -> None:

    init(default_config.CONFIG_FILE_PATH)
    con_dir, def_file = def_settings_dict.get('CONFIG_DIR'), f"{receipe}.toml"
    con_dir_full = os.path.join(os.getcwd(), con_dir)
    def_file_full = os.path.join(con_dir_full, def_file)

    if not check_if_receipe_exists(receipe):
        logging.warning(
            f"receipe {receipe} doesn't exist in the config directory")
        logging.warning("use `scm create` for the receipe creation")
        raise typer.Exit()

    res_validation = validate_unsupported_resources(get_user_settings(receipe))

    if res_validation:
        logging.warning(
            f"Unsupported resources found in the `{receipe}` configuration file"
        )
        for i in res_validation:
            logging.warning(
                f"`{i}` resource in {receipe} receipe isn't supported")
            raise typer.Exit()

    user_settings = get_user_settings(receipe)
    user_resources = get_user_defined_resources(user_settings)

    if not user_resources:
        logging.warning(
            f"No user resources found in the `{receipe}` configuration file")
        raise typer.Exit()

    for res in user_resources:
        if res in user_settings:
            for i in user_settings[res]:

                if not isinstance(user_settings[res][i], DynaBox):
                    logging.warning(
                        f"Missing subconfig `{res}` resource in {receipe} receipe"
                    )
                    raise typer.Exit()

                if not user_settings[res][i].get('name', None):
                    logging.warning(
                        f"Missing `name`attributes in resource `{i}` in receipe {receipe}"
                    )
                    raise typer.Exit()

                if not user_settings[res][i].get('action', None):
                    logging.warning(
                        f"Missing `action` attributes in resource `{i}` in receipe {receipe}"
                    )
                    raise typer.Exit()

                defaults.UNSUPPORTED_ATTR = OrderedSet(
                    user_settings[res][i].keys()) - defaults.RES_ATTRIBUTES

                if (defaults.UNSUPPORTED_ATTR -
                        defaults.SRV_ATTRIBUTES) and res.upper() == "SERVICE":
                    logging.warning(
                        f"Found one more attributes that aren't supported"
                    )
                    for unsup in (
                            defaults.UNSUPPORTED_ATTR -
                            defaults.SRV_ATTRIBUTES):
                        logging.warning(
                            f"`{unsup}` not supported attribute in resource {res} in receipe {receipe}"
                        )
                    raise typer.Exit()

                if (defaults.UNSUPPORTED_ATTR -
                        defaults.DIR_ATTRIBUTES) and res.upper() == "DIRECTORY":
                    logging.warning(
                        f"Found one more attributes that aren't supported"
                    )
                    for unsup in (
                            defaults.UNSUPPORTED_ATTR -
                            defaults.DIR_ATTRIBUTES):
                        logging.warning(
                            f"`{unsup}` not supported attribute in resource {res} in receipe {receipe}"
                        )
                    raise typer.Exit()

                if (defaults.UNSUPPORTED_ATTR -
                        defaults.FILE_ATTRIBUTES) and res.upper() == "FILE":
                    logging.warning(
                        f"Found one more attributes that aren't supported"
                    )
                    for unsup in (
                            defaults.UNSUPPORTED_ATTR -
                            defaults.FILE_ATTRIBUTES):
                        logging.warning(
                            f"`{unsup}` not supported attribute in resource {res} in receipe {receipe}"
                        )
                    raise typer.Exit()

                # now validate the values from the user-input string
                user_values = user_settings[res][i].get('action', None)
                if res.upper() == "SERVICE":
                    for val in user_values:
                        if val not in defaults.SERVICE_SETUP_ACTIONS.union(
                                defaults.SERVICE_OP_ACTIONS):
                            logging.warning(
                                f"`{val}` not supported in action attribute in resource {res} in receipe {receipe}"
                            )
                            raise typer.Exit()
                
                else:
                    if res.upper() in ["FILE", "DIRECTORY"]:
                        for val in user_values:
                            if val not in defaults.DIR_FILE_ACTIONS:
                                logging.warning(
                                    f"`{val}` not supported in action attribute in resource {res} in receipe {receipe}"
                                )
                                raise typer.Exit()
                
                    if res.upper() in ["FIREWALL"]:
                        for val in user_values:
                            if val not in defaults.FIREWALL_ACTIONS:
                                logging.warning(
                                    f"`{val}` not supported in action attribute in resource {res} in receipe {receipe}"
                                )
                                raise typer.Exit()

    logging.info(
        f"{receipe} receipe file is valid for push, use `scm diff` to differences with the existing configuration")
    return 0

def push_command(receipe) -> tuple:
    
    curr_resouces = OrderedDict()

    settings = read_json(default_config.CONFIG_FILE_PATH)['default']

    os.environ['ROOT_PATH_FOR_DYNACONF'] = settings.get(
        'CONFIG_DIR', os.path.join(os.getcwd(), "config"))

    hash_config_dir = os.path.join(
        settings.get('CONFIG_HASH_DIR'),
        settings.get('CONFIG_HASH_FILE'))

    validate(receipe)

    user_settings = get_user_settings(receipe)
    user_resources = get_user_defined_resources(user_settings)
    
    curr_resouces[receipe] = {}   
    for i in user_resources:
        gen_command(user_settings, i, curr_resouces[receipe])
    
    write_dict = {}
    write_dict[receipe] = {}
    
    for index,value in curr_resouces[receipe].items():
        concat = functools.reduce(lambda x, y: x + y, value, "")
        write_dict[receipe][index] = hashlib.md5(concat.encode("utf-8")).hexdigest()
    
    existing_hash = read_json(hash_config_dir)
    
    if not existing_hash:
        existing_hash[receipe] = {}
    
    diff_output = _get_diff_hash(existing_hash[receipe],write_dict[receipe])
      
    return (diff_output, curr_resouces, write_dict)


@app.command()
def push(
    receipe: str = typer.Option(...)
) -> None:
    diff_output, curr_resources, new_hash_dict = push_command(receipe)
    
    if not diff_output:
        logging.info(f"`{receipe}` configuration is update to date with the existing configuration")
        raise typer.Exit()
    
    
    settings = read_json(default_config.CONFIG_FILE_PATH)['default']
    
    hash_config_dir = os.path.join(
        settings.get('CONFIG_HASH_DIR'),
        settings.get('CONFIG_HASH_FILE'))
    
    logging.info("Following resources will be applied:")
    for cmd in diff_output: 
        if cmd in  curr_resources[receipe]:       
            logging.info(f"{cmd}: {curr_resources[receipe][cmd]}")
            for c in curr_resources[receipe][cmd]: 
                logging.info(f"Applying the command `{c}`")
                code = run_os_command(c)
                if code:
                    logging.error(f"Failed to run the command {c}")
                    raise typer.Exit()
                
    with open(hash_config_dir, "w") as output:
        json.dump(new_hash_dict, output)

    return 0


@app.command()
def diff(
    receipe: str = typer.Option(...)
) -> None:
    
    output, curr_resources, new_hash_dict = push_command(receipe)
    
    if not output:
        logging.info(f"`{receipe}` configuration is update to date with the existing configuration")
    
    for i in output: 
        if i in  curr_resources[receipe]:
            logging.info("Following resources will be applied:")
            logging.info(f"{i}: {curr_resources[receipe][i]}")    
        
    return 0


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        help="Display's the application version",
        callback=_version,
        is_eager=True,
    )
) -> None:
    return 0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scm_config']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.4,<0.5.0',
 'deepdiff>=5.7.0,<6.0.0',
 'dynaconf>=3.1.5,<4.0.0',
 'ordered-set>=4.0.2,<5.0.0',
 'pytest>=7.0.1,<8.0.0',
 'typer>=0.4.1,<0.5.0',
 'wheel>=0.37.1,<0.38.0']

entry_points = \
{'console_scripts': ['scm = scm_config.main:main_cli']}

setup_kwargs = {
    'name': 'scm-config',
    'version': '0.2.4',
    'description': 'self configuration management written in Python',
    'long_description': '# SCM\nSelf Managed Configuration Management(SCM) is a simple command line app for configuration management, written in python.\n> Designed to run only on ubuntu operating system\n\nThis tool, currently supports only below resources \n* service\n* directory\n* file \n* firewall\n\n\n# Resource Detailed information \n## Service \nService resource is useful for installing and managing packages from the linux repository, \n> Note that currently this tool is designed to support only on Ubuntu Operating System\n\n***Name parameter in the below section is a list format, meaning it can take multiple values and the actions and other parameters are applied to each name resource accordingly**\n### Example \n```\n[service.setup]\nname = ["apache2"] \naction= ["install", "enable"]\n```\nIn the above example, "service" is the resource and "setup" is the service identifier. \n* name  -> Name of the service that is to be installed on the ubuntu system \n* action -> Action to be done on the listed service, currently these are only supported \n> ["install", "enable", "disable"]\n\n```\n[service.ops]\nname = ["apache2"]\naction =["restart"]\n```\nIn the above example, "service" is the resource and "ops" is the service identifier. once the service is installed and service operations can be done using this tool. \n* name  -> Name of the service that is to be installed on the ubuntu system \n* action -> Action to be done on the listed service, currently these are only supported \n> ["stop", "start", "restart", "reload", "disable", "enable"]\n\n## Directory\nDirectory resource is useful for installing and managing the metadata directories on the system, \n> Note that currently this resource is tested only on Ubuntu Operating System, but can run on any Linux Operating System\n\n### Example \n```\n# Modify the directory permissions \n[directory.list]\nname = ["/etc/apache"] \nparams = {\'owner\'=\'root\',\'group\'=\'root\',\'mode\'= \'0755\'}\naction = [\'create\']\nnotifies = "@format {this.service.ops}"\n```\nIn the above example, "directory" is the resource and "list" is the service identifier. \n* name  -> Name of the directory where the metadata needs modification including creation of the directory using the action method\n* params -> parameter to support the directory operations, currently [\'owner\', \'group\' and \'mode\'] are supported in the list \n* action -> Action to be done on the listed service, currently for directories only [\'create\'] is supported\n* notifies -> Parameter to notify any other resource in the same recipe file, In example, notifies the service ops resources, that would restart the apache2 service based on the modifications \n> ["install", "enable", "disable"]\n\n### Example \n```\n# Modify the input content of the file \n[file.conf]\nname = ["/var/www/customers/public_html/index.php"]\naction = ["create"]\noverride ="true"\ncontent  = ["This is the test file"]\nparams = {\'owner\'= \'root\',\'group\'= \'root\',\'mode\' = \'0755\'}\nnotifies = "@format {this.service.ops}"\n\n```\nIn the above example, "file" is the resource and "conf" is the service identifier. \n* name  -> Name of the file where the content need to be added or appended based on the configuration requirement\n* action -> Action to be done on the listed service, currently for file only [\'create\'] is supported\n* override -> This parameter will override if there is any existing file, default it will append the content to the file \n* content -> Input content for the file provide in the form the double quotes. For simplicity ,large content is not tested with the current version of the code. \n* params -> parameter to support the file operations, currently [\'owner\', \'group\' and \'mode\'] are supported in the list \n* notifies -> Parameter to notify any other resource in the same recipe file, In example, notifies the service ops resources, that would restart the apache2 service based on the modifications \n> ["install", "enable", "disable"]\n\n# Complete overview of the example file \n\n```toml\n# Service setup for the apache instance \n[service.setup]\nname = ["apache2"] \naction= ["install", "enable"]\n\n# Service restart for the apache instance \n[service.ops]\nname = ["apache2"]\naction =["restart"]\n\n\n# Modify the directory permissions \n[directory.list]\nname = ["/etc/apache"] \nparams = {\'owner\'=\'root\',\'group\'=\'root\',\'mode\'= \'0755\'}\naction = [\'create\']\nnotifies = "@format {this.service.ops}"\n\n# Modify the input content of the file \n[file.conf]\nname = ["/var/www/customers/public_html/index.php"]\naction = ["create"]\ncontent  = ["This is the test file"]\nparams = {\'owner\'= \'root\',\'group\'= \'root\',\'mode\' = \'0755\'}\nnotifies = "@format {this.service.ops}"\n```\n# Installation and Usage # \n\n## Manual clone\n```bash\n  $ git clone https://github.com/Sai-Repalle/scm_project\n  $ cd scm_project\n  $ python3 -m venv venv \n  $ source venv/bin/activate\n  $ pip install -r requirements.txt \n```\n## Usage\n```bash\nUsage: scm [OPTIONS] COMMAND [ARGS]...\nOptions:\n  -v, --version         Display\\\'s the application version\n  --install-completion  Install completion for the current shell.\n  --show-completion     Show completion for the current shell, to copy it or\n                        customize the installation.\n  --help                Show this message and exit.\n\nCommands:\n  create\n  diff\n  info\n  init\n  push\n  validate\n\n```\n## initialization\ninit command is used for initialization and helpful to create the respective directories, also this command is useful for expanding future versions of the scm tool\n\n`init`\nTo start with the tool, initialize the tool without any parameters, this would set the configuration files required for tool to work\n```bash\n$ scm init\n```\n### output \n```bash\nscm init   \nroot@machine:~/scm# python3 -m scm init\n[INFO][04-22-2022 06:08:13]::Reading the Json configuration /root/scm/scm/settings/settings.json\n[INFO][04-22-2022 06:08:13]::creating directory CONFIG_DIR\n[INFO][04-22-2022 06:08:13]::creating directory CONFIG_HASH_DIR\n[INFO][04-22-2022 06:08:13]::creating files CONFIG_DEF_FILE\n[INFO][04-22-2022 06:08:13]::creating files CONFIG_HASH_FILE\n\n```\n## create\n`create <name>`\n```bash\n$ scm init --receipe_name <receipe_name>\n```\nBelow example, will create a receipe called "apache" and `apache.toml` file is created in the config directory located at the root directory of the scm \n### output \n```bash\nroot@machine:~/scm# python3 -m scm create --receipe apache\n[INFO][04-22-2022 06:09:56]::Reading the Json configuration /root/scm/scm/settings/settings.json\n[INFO][04-22-2022 06:09:56]::creating directory CONFIG_DIR\n[INFO][04-22-2022 06:09:56]::CONFIG_DIR directory already exists\n[INFO][04-22-2022 06:09:56]::creating directory CONFIG_HASH_DIR\n[INFO][04-22-2022 06:09:56]::CONFIG_HASH_DIR directory already exists\n[INFO][04-22-2022 06:09:56]::creating files CONFIG_DEF_FILE\n[INFO][04-22-2022 06:09:56]::creating files CONFIG_HASH_FIL\n```\n\n\n## info\n`info --receipe_name <name>`\n```bash\n$ scm info --receipe_name <receipe_name>\n```\n```bash \n\n```\n## validate\n`validate <name>`\n```bash\n$ scm validate --receipe_name <receipe_name>\n```\n## diff\n`diff <name>`\n```bash\n$ scm diff --receipe_name <receipe_name>\n```\n## push\n`push <name>`\n```bash\n$ scm push --receipe_name <receipe_name>\n```',
    'author': 'Sai Repalle',
    'author_email': 'n.repalle85@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Sai-Repalle/scm_project',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import os
import logging
from typing import List, Set, Dict
from dynaconf import Dynaconf
from dynaconf.utils.boxing import DynaBox
from dynaconf.vendor.toml.decoder import TomlDecodeError
from ordered_set import OrderedSet
import json
import hashlib
from deepdiff import DeepDiff
from deepdiff.model import PrettyOrderedSet

logging.basicConfig(format='[%(levelname)s][%(asctime)s]::%(message)s',
                    datefmt="%m-%d-%Y %H:%M:%S")
logger = logging.getLogger()
supp_res = OrderedSet(['SERVICE', 'FILE', 'DIRECTORY'])
service_setup_actions = OrderedSet(["install", "enable", "disable"])
service_op_actions = OrderedSet(
    ["stop", "start", "restart", "reload", "disable", "enable"])
dir_file_actions = OrderedSet(['create'])
res_attributes = OrderedSet(['name', 'action'])
srv_attributes = OrderedSet(['name', 'action', 'notifies'])
dir_attributes = OrderedSet(['params', 'notifies'])
file_attributes = OrderedSet(['params', 'notifies', 'content'])
json_diff_attr = OrderedSet(
    ['dictionary_item_added', 'dictionary_item_removed', 'values_changed'])


os.environ['ROOT_PATH_FOR_DYNACONF'] = "./my_configuration"
hash_file = "hash_config.md5.json"
hash_configuration = os.path.join(
    os.environ['ROOT_PATH_FOR_DYNACONF'], hash_file)


default_paramters = OrderedSet([
    'SETTINGS_FILE_FOR_DYNACONF', 'RENAMED_VARS', 'ROOT_PATH_FOR_DYNACONF',
    'ENVIRONMENTS_FOR_DYNACONF', 'MAIN_ENV_FOR_DYNACONF',
    'LOWERCASE_READ_FOR_DYNACONF', 'ENV_SWITCHER_FOR_DYNACONF',
    'ENV_FOR_DYNACONF', 'FORCE_ENV_FOR_DYNACONF', 'DEFAULT_ENV_FOR_DYNACONF',
    'ENVVAR_PREFIX_FOR_DYNACONF', 'IGNORE_UNKNOWN_ENVVARS_FOR_DYNACONF',
    'ENCODING_FOR_DYNACONF', 'MERGE_ENABLED_FOR_DYNACONF',
    'NESTED_SEPARATOR_FOR_DYNACONF', 'ENVVAR_FOR_DYNACONF',
    'REDIS_FOR_DYNACONF', 'REDIS_ENABLED_FOR_DYNACONF', 'VAULT_FOR_DYNACONF',
    'VAULT_ENABLED_FOR_DYNACONF', 'VAULT_PATH_FOR_DYNACONF',
    'VAULT_MOUNT_POINT_FOR_DYNACONF', 'VAULT_ROOT_TOKEN_FOR_DYNACONF',
    'VAULT_KV_VERSION_FOR_DYNACONF', 'VAULT_AUTH_WITH_IAM_FOR_DYNACONF',
    'VAULT_AUTH_ROLE_FOR_DYNACONF', 'VAULT_ROLE_ID_FOR_DYNACONF',
    'VAULT_SECRET_ID_FOR_DYNACONF', 'CORE_LOADERS_FOR_DYNACONF',
    'LOADERS_FOR_DYNACONF', 'SILENT_ERRORS_FOR_DYNACONF',
    'FRESH_VARS_FOR_DYNACONF', 'DOTENV_PATH_FOR_DYNACONF',
    'DOTENV_VERBOSE_FOR_DYNACONF', 'DOTENV_OVERRIDE_FOR_DYNACONF',
    'INSTANCE_FOR_DYNACONF', 'YAML_LOADER_FOR_DYNACONF',
    'COMMENTJSON_ENABLED_FOR_DYNACONF', 'SECRETS_FOR_DYNACONF',
    'INCLUDES_FOR_DYNACONF', 'PRELOAD_FOR_DYNACONF', 'SKIP_FILES_FOR_DYNACONF',
    'DYNACONF_NAMESPACE', 'NAMESPACE_FOR_DYNACONF', 'DYNACONF_SETTINGS_MODULE',
    'DYNACONF_SETTINGS', 'SETTINGS_MODULE', 'SETTINGS_MODULE_FOR_DYNACONF',
    'PROJECT_ROOT', 'PROJECT_ROOT_FOR_DYNACONF', 'DYNACONF_SILENT_ERRORS',
    'DYNACONF_ALWAYS_FRESH_VARS', 'BASE_NAMESPACE_FOR_DYNACONF',
    'GLOBAL_ENV_FOR_DYNACONF'])


def check_if_receipe_exists(receipe) -> bool:
    return os.path.exists(
        os.path.join(os.getcwd(), os.environ['ROOT_PATH_FOR_DYNACONF'],
                     f"{receipe}.toml"))


def validate_unsupported_resources(user_resources) -> Set:
    global default_paramters, supp_res
    unsupported_resources = ({*user_resources} - default_paramters) - supp_res
    return unsupported_resources


def get_user_settings(receipe, validator=None, environments=True) -> Dict:
    return Dynaconf(settings_files=[f"{receipe}.toml"], validators=validator)


def get_user_defined_resources(settings) -> Set:
    return (OrderedSet([*settings]) - default_paramters)


def create(receipe) -> None:
    """creates the configuration file based on the input"""
    con_dir, def_file = os.environ['ROOT_PATH_FOR_DYNACONF'], f"{receipe}.toml"
    con_dir_full = os.path.join(os.getcwd(), con_dir)
    def_file_full = os.path.join(con_dir_full, def_file)

    if not os.path.isdir(con_dir_full):
        os.mkdir(con_dir_full)

    if not check_if_receipe_exists(receipe):
        with open(def_file_full, mode="w") as f:
            f.write("# This is the chef receipe that I created the file")
    else:
        logging.warning(
            f"`{receipe}` configuration file exists, use --force to override.")


def _hash_fun(user_dict: Dict) -> str:
    user_dict = json.dumps(user_dict, sort_keys=True)
    return hashlib.md5(user_dict.encode('utf-8')).hexdigest()


def _write_json(new_data, filename) -> None:
    with open(filename, 'w') as file:
        filedata = json.load(f)
        filedata.update(new_data)
        file.seek(0)
        json.dump(filedata, file, indent=4)


def _read_json(filename) -> Dict:
    f = open(filename)
    try:
        dict_output = json.load(f)
    except Exception as e:
        return None


def _get_diff_hash(existing_hash, curr_hash, receipe) -> List:
    output = []
    if existing_hash:
        res = DeepDiff(curr_hash, existing_hash)
        for i in json_diff_attr:
            if res.get(i, None):
                if isinstance(res.get(i, None), PrettyOrderedSet):
                    for val in res.get(i, None):
                        split_val = val.split('[0]')[1]
                        if split_val[0] == '[' and split_val[-1] == "]":
                            split_val = split_val[1:-1]
                        output.append(split_val.strip("\'"))

    else:
        for i in curr_hash[receipe]:
            output += list(i.keys())
    return output


def get_diff_hash_res(receipe) -> List:

    user_settings = dict(get_user_settings(receipe))
    user_resources = get_user_defined_resources(user_settings)
    data = {f"{receipe}": []}
    hash_dict = {}

    existing_hash = _read_json(hash_configuration)

    for res in user_resources:
        if res in user_settings:
            for i, value in user_settings[res].items():
                hash_val = _hash_fun(user_settings[res][i])
                hash_dict[f"{res}.{i}"] = hash_val

    data[f"{receipe}"] = [hash_dict]

    diff_resources = _get_diff_hash(existing_hash, data, receipe)

    return diff_resources


# write configuration to the build hash set
def write_hash_config(hash_dict_set, receipe, hash_file_name):
    output = {}
    user_settings = dict(get_user_settings(receipe))
    user_resources = get_user_defined_resources(user_settings)

    for res in user_resources:
        if res in user_settings:
            for key, val in user_settings[res].items():
                valid_con = f"{res}.{key}"
                if valid_con in hash_dict_set:
                    output[key] = dict(val)
    return {receipe: output}


def info(receipe) -> None:
    """list out the info for the settings"""
    # if validate(receipe):
    #     settings = get_user_settings(receipe)
    pass


def validate(receipe, settings) -> bool:
    """Validate the user configuration"""

    if not check_if_receipe_exists(receipe):
        logging.error(f"`{i}` receipe doesn't exist in configuration files")
        return False

    res_validation = validate_unsupported_resources(get_user_settings(receipe))

    if res_validation:
        logging.warning(
            f"Unsupported resources found in the `{receipe}` configuration file"
        )
        for i in res_validation:
            logging.warning(
                f"`{i}` resource in {receipe} receipe isn't supported")
            return False

    user_settings = get_user_settings(receipe)
    user_resources = get_user_defined_resources(user_settings)

    if not user_resources:
        logging.warning(
            f"No user resources found in the `{receipe}` configuration file")
        return False

    for res in user_resources:
        if res in user_settings:
            for i in user_settings[res]:

                if not isinstance(user_settings[res][i], DynaBox):
                    logging.warning(
                        f"Missing subconfig `{res}` resource in {receipe} receipe"
                    )
                    return False

                if not user_settings[res][i].get('name', None):
                    logging.warning(
                        f"Missing `name`attributes in resource `{i}` in receipe {receipe}"
                    )
                    return False

                if not user_settings[res][i].get('action', None):
                    logging.warning(
                        f"Missing `action` attributes in resource `{i}` in receipe {receipe}"
                    )
                    return False

                unsupported_attr = OrderedSet(
                    user_settings[res][i].keys()) - res_attributes

                if (unsupported_attr - srv_attributes) and res.upper() == "SERVICE":
                    logging.warning(
                        f"Found one more attributes that aren't supported"
                    )
                    for unsup in (unsupported_attr - srv_attributes):
                        logging.warning(
                            f"`{unsup}` not supported attribute in resource {res} in receipe {receipe}"
                        )
                    return False

                if (unsupported_attr - dir_attributes) and res.upper() == "DIRECTORY":
                    logging.warning(
                        f"Found one more attributes that aren't supported"
                    )
                    for unsup in (unsupported_attr - dir_attributes):
                        logging.warning(
                            f"`{unsup}` not supported attribute in resource {res} in receipe {receipe}"
                        )
                    return False

                if (unsupported_attr - file_attributes) and res.upper() == "FILE":
                    logging.warning(
                        f"Found one more attributes that aren't supported"
                    )
                    for unsup in (unsupported_attr - file_attributes):
                        logging.warning(
                            f"`{unsup}` not supported attribute in resource {res} in receipe {receipe}"
                        )
                    return False

                # now validate the values from the user-input string
                user_values = user_settings[res][i].get('action', None)
                if res.upper() == "SERVICE":
                    for val in user_values:
                        if val not in service_setup_actions.union(
                                service_op_actions):
                            logging.warning(
                                f"`{val}` not supported in action attribute in resource {res} in receipe {receipe}"
                            )

                else:
                    if res.upper() in ["FILE", "DIRECTORY"]:
                        for val in user_values:
                            if val not in dir_file_actions:
                                logging.warning(
                                    f"`{val}` not supported in action attribute in resource {res} in receipe {receipe}"
                                )
                print(user_settings[res][i])

    return True


def gen_command(settings_dict: Dict, key: str, output: List) -> None:
    """Function to the generate the OS commands by reading the settings_dict and update the input list format"""
    if key.upper() in ["SERVICE"]:
        for index, value in settings_dict[key].items():
            if isinstance(value, DynaBox) and index != "params":
                for n in value['name']:
                    for act in value['action']:
                        if act in service_setup_actions:
                            output.append("sudo apt update")
                            output.append(f"sudo apt {act} {n} -y")
                        else:
                            if act in service_op_actions:
                                output.append(f"systemctl {act} {n} -force")

    if key.upper() in ["DIRECTORY", "FILE"]:
        for index, value in settings_dict[key].items():
            for n in value['name']:
                for act in value['action']:
                    if act == "create" and value['params'].get(
                            'owner', None) and value['params'].get('group', None):
                        cmd: str = f"chown {value['params']['owner']}:{value['params']['group']} {n}"
                        if value['params'].get('recurse', None) and json.loads(
                                (value['params'].get('recurse', None)).lower()):
                            cmd += " -R"
                        output.append(cmd)

            if value.get('notifies', None):
                inp_json = json.loads(
                    value.get(
                        'notifies',
                        None).replace(
                        "\'",
                        "\""))
                for n in inp_json['name']:
                    for act in inp_json['action']:
                        if act == "install":
                            output.append(f"apt-get {act} {n} -y")
                        else:
                            output.append(f"systemctl {act} {n} -force")

    return None


def push(receipe) -> None:
    """Pushes the configuration to the operating system"""
    output = []
    user_settings = get_user_settings(receipe)
    user_resources = get_user_defined_resources(user_settings)

    for key, value in user_settings.items():
        if key in user_resources:
            res = gen_command(user_settings, key, output)

    print(output)


def clean(receipe) -> None:
    pass

import subprocess 

def loop_through_os_commands(input) -> None:
    for i in input: 
        process = subprocess.Popen(i.split(), stdout = subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        
        while True:
            output = process.stdout.readline()
            print(output.strip())
            # Do something else
            return_code = process.poll()
            if return_code is not None:
                # Process has finished, read rest of the output 
                for output in process.stdout.readlines():
                    print(output.strip())
                break
    



if __name__ == "__main__":
    loop_through_os_commands(['powershell.exe echo hello world', 'powershell.exe echo test1'])
    



# default configuration for the code
from ordered_set import OrderedSet

SUPP_RES = OrderedSet(['SERVICE', 'FILE', 'DIRECTORY','FIREWALL'])
SERVICE_SETUP_ACTIONS = OrderedSet(["install", "enable", "disable"])
SERVICE_OP_ACTIONS = OrderedSet(
    ["stop", "start", "restart", "reload", "disable", "enable"])
DIR_FILE_ACTIONS = OrderedSet(['create'])
FIREWALL_ACTIONS = OrderedSet(['allow'])
RES_ATTRIBUTES = OrderedSet(['name', 'action'])
SRV_ATTRIBUTES = OrderedSet(['name', 'action', 'notifies'])
DIR_ATTRIBUTES = OrderedSet(['params', 'notifies'])
FILE_ATTRIBUTES = OrderedSet(['params', 'notifies', 'content'])
JSON_DIFF_ATTR = OrderedSet(
    ['dictionary_item_added', 'dictionary_item_removed', 'values_changed'])
DEFAULT_PARAMTERS = OrderedSet([
    'SETTINGS_FILE_FOR_DYNACONF', 'RENAMED_VARS', 'ROOT_PATH_FOR_DYNACONF',
    'ENVIRONMENTS_FOR_DYNACONF', 'MAIN_ENV_FOR_DYNACONF',
    'LOWERCASE_READ_FOR_DYNACONF', 'ENV_SWITCHER_FOR_DYNACONF',
    'ENV_FOR_DYNACONF', 'FORCE_ENV_FOR_DYNACONF', 'DEFAULT_ENV_FOR_DYNACONF',
    'ENVVAR_PREFIX_FOR_DYNACONF', 'IGNORE_UNKNOWN_ENVVARS_FOR_DYNACONF',
    'ENCODING_FOR_DYNACONF', 'MERGE_ENABLED_FOR_DYNACONF',
    'NESTED_SEPARATOR_FOR_DYNACONF', 'ENVVAR_FOR_DYNACONF',
    'REDIS_FOR_DYNACONF', 'REDIS_ENABLED_FOR_DYNACONF', 'VAULT_FOR_DYNACONF',
    'VAULT_ENABLED_FOR_DYNACONF', 'VAULT_PATH_FOR_DYNACONF',
    'VAULT_MOUNT_POINT_FOR_DYNACONF', 'VAULT_ROOT_TOKEN_FOR_DYNACONF',
    'VAULT_KV_VERSION_FOR_DYNACONF', 'VAULT_AUTH_WITH_IAM_FOR_DYNACONF',
    'VAULT_AUTH_ROLE_FOR_DYNACONF', 'VAULT_ROLE_ID_FOR_DYNACONF',
    'VAULT_SECRET_ID_FOR_DYNACONF', 'CORE_LOADERS_FOR_DYNACONF',
    'LOADERS_FOR_DYNACONF', 'SILENT_ERRORS_FOR_DYNACONF',
    'FRESH_VARS_FOR_DYNACONF', 'DOTENV_PATH_FOR_DYNACONF',
    'DOTENV_VERBOSE_FOR_DYNACONF', 'DOTENV_OVERRIDE_FOR_DYNACONF',
    'INSTANCE_FOR_DYNACONF', 'YAML_LOADER_FOR_DYNACONF',
    'COMMENTJSON_ENABLED_FOR_DYNACONF', 'SECRETS_FOR_DYNACONF',
    'INCLUDES_FOR_DYNACONF', 'PRELOAD_FOR_DYNACONF', 'SKIP_FILES_FOR_DYNACONF',
    'DYNACONF_NAMESPACE', 'NAMESPACE_FOR_DYNACONF', 'DYNACONF_SETTINGS_MODULE',
    'DYNACONF_SETTINGS', 'SETTINGS_MODULE', 'SETTINGS_MODULE_FOR_DYNACONF',
    'PROJECT_ROOT', 'PROJECT_ROOT_FOR_DYNACONF', 'DYNACONF_SILENT_ERRORS',
    'DYNACONF_ALWAYS_FRESH_VARS', 'BASE_NAMESPACE_FOR_DYNACONF',
    'GLOBAL_ENV_FOR_DYNACONF'])

SETTINGS_DATA = {
            "default": {
                "DEBUG": False,
                "CONFIG_DIR": "config",
                "CONFIG_DEF_FILE": "default.toml",
                "CONFIG_HASH_DIR": "config_hash",
                "CONFIG_HASH_FILE": "hash_config_md5.json"
            }}


"""Default configuration settings for the dynaconf"""
# scm/defaults_config.py
import os
import logging
from typing import List, Set, Dict
from dynaconf import Dynaconf
from dynaconf.utils.boxing import DynaBox
from dynaconf.vendor.toml.decoder import TomlDecodeError
from ordered_set import OrderedSet
import json
import hashlib
import shlex 
from deepdiff import DeepDiff
from deepdiff.model import PrettyOrderedSet
from json.decoder import JSONDecodeError
from scm_config.defaults import JSON_DIFF_ATTR
from pathlib import Path
from scm_config import defaults
from subprocess import STDOUT, check_call, CalledProcessError

from scm_config import (
    DIR_ERROR,
    FILE_ERROR,
    SUCCESS,
    __app_name__

)

CONFIG_DIR = OrderedSet(['CONFIG_DIR', 'CONFIG_HASH_DIR'])
CONFIG_FILES = OrderedSet(['CONFIG_DEF_FILE', 'CONFIG_HASH_FILE'])
CONFIG_DIR_PATH = os.getcwd()
CONFIG_FILE_PATH = os.path.join(
    CONFIG_DIR_PATH,
    "settings.json")


def read_json(filename):
    try:
        with open(filename) as file:
            logging.info(f"Reading the Json configuration {filename}")
            try:
                return dict(json.load(file))
            except JSONDecodeError:
                logging.warning(f"Invalid JSON file {file}")
                return FILE_ERROR
    except FileNotFoundError:
        logging.ERROR(f"File {filename} not found")
        return FILE_ERROR


def create_def_directory(user_dict) -> bool:
    for key in CONFIG_DIR:
        logging.info(f"creating directory {key}")
        if user_dict.get(key, None):
            if not os.path.exists(user_dict[key]):
                os.mkdir(user_dict[key])
            else:
                logging.info(f"{key} directory already exists")
        else:
            logging.warning(f"Missing {key} value to create the directories")
            return False
    return True


def create_def_files(user_dict) -> bool:
    for key in CONFIG_FILES:
        logging.info(f"creating files {key}")
        if user_dict.get(key, None):
            if not os.path.exists(user_dict[key]):
                if key == "CONFIG_DEF_FILE":
                    Path(
                        os.path.join(
                            user_dict['CONFIG_DIR'],
                            user_dict[key])).touch()
                else:
                    hash_directory = os.path.join(
                            user_dict['CONFIG_HASH_DIR'],
                            user_dict[key])
                    if not os.path.exists(hash_directory):
                        Path(hash_directory).touch()
                        #write empty json dict to the file 
                        d = {}
                        # Serializing json 
                        json_object = json.dumps(d, indent = 4)
                        # Writing to sample.json
                        with open(os.path.join(
                                user_dict['CONFIG_HASH_DIR'],
                                user_dict[key]), "w") as outfile:
                            outfile.write(json_object)
            else:
                logging.info(f"{key} file already exists")
        else:
            logging.warning(f"Missing {key} value to create the files")
            return False

    return True


def check_if_receipe_exists(receipe) -> bool:
    return os.path.exists(
        os.path.join(os.getcwd(), os.environ['ROOT_PATH_FOR_DYNACONF'],
                     f"{receipe}.toml"))


def get_user_settings(receipe, validator=None, environments=True) -> Dict:
    return Dynaconf(settings_files=[f"{receipe}.toml"], validators=validator)


def get_user_defined_resources(settings) -> Set:
    return (OrderedSet([*settings]) - defaults.DEFAULT_PARAMTERS)


def validate_unsupported_resources(user_resources) -> Set:
    unsupported_resources = (
        {*user_resources} - defaults.DEFAULT_PARAMTERS) - defaults.SUPP_RES
    return unsupported_resources


def gen_command(
        settings_dict: Dict,
        key: str,
        output: List) -> None:
    """Function to the generate the OS commands by reading the settings_dict and update the input list format"""
    if key.upper() in ["SERVICE"]:
        for index, value in settings_dict[key].items():
            output[f"{key}.{index}"] = []
            if isinstance(
                    value, DynaBox) and index != "params":
                for n in value['name']:
                    for act in value['action']:
                        if act in defaults.SERVICE_SETUP_ACTIONS:
                            output[f"{key}.{index}"].append("sudo apt-get update -y")
                            output[f"{key}.{index}"].append(f"sudo apt-get {act} {n} -y")
                        else:
                            if act in defaults.SERVICE_OP_ACTIONS:
                                output[f"{key}.{index}"].append(f"systemctl {act} {n} -force")

    if key.upper() in ["DIRECTORY", "FILE"]:
        for index, value in settings_dict[key].items():
            output[f"{key}.{index}"] = []
            for n in value['name']:
                for act in value['action']:
                    if act == "create" and value.get('content',None):
                        for i in value['content']:
                            if not value.get('override',None):
                                cmd: str = f"echo '{i}' >> {n}"
                            else:
                                cmd:str = f"echo '{i}' > {n}"
                            output[f"{key}.{index}"].append(cmd)

                    if act == "create" and key == "DIRECTORY":
                        cmd: str = f"mkdir -p {n}"
                        output[f"{key}.{index}"].append(cmd)

                    if act == "create" and value['params'].get(
                            'owner',
                            None) and value['params'].get(
                            'group',
                            None):
                        cmd: str = f"chown {value['params']['owner']}:{value['params']['group']} {n}"
                        if value['params'].get('recurse', None) and json.loads(
                                (value['params'].get('recurse', None)).lower()):
                            cmd += " -R"

                        output[f"{key}.{index}"].append(cmd)


                if value.get('notifies', None):
                    inp_json = json.loads(
                        value.get(
                            'notifies',
                            None).replace(
                            "\'",
                            "\""))
                    for n in inp_json['name']:
                        for act in inp_json['action']:
                            if act == "install":
                                output[f"{key}.{index}"].append(f"apt-get {act} {n} -y")
                            else:
                                output[f"{key}.{index}"].append(f"systemctl {act} {n} -force")

    if key.upper() in ["FIREWALL"]:
        for index, value in settings_dict[key].items():
            output[f"{key}.{index}"] = []
            for n in value['name']:
                for act in value['action']:
                    if act == "allow":
                        cmd: str = f"sudo ufw {act} in {n}"
                        output[f"{key}.{index}"].append(cmd)
                        
                        
    return None



def _hash_fun(user_dict: Dict) -> str:
    user_dict = json.dumps(user_dict, sort_keys=True)
    return hashlib.md5(user_dict.encode('utf-8')).hexdigest()


def _write_json(new_data, filename) -> None:
    with open(filename, 'w') as file:
        filedata = json.load(new_data)
        print(filedata)
        filedata.update(new_data)
        file.seek(0)
        json.dump(filedata, file, indent=4)


def _read_json(filename) -> Dict:
    f = open(filename)
    try:
        dict_output = json.load(f)
    except Exception as e:
        return None
    return dict_output 

def _get_diff_hash(existing_hash, curr_hash) -> List:
    output = []
    if existing_hash:
        res = DeepDiff(curr_hash, existing_hash)
        for i in JSON_DIFF_ATTR:
            if res.get(i, None):
                for split_val in res.get(i, None):
                    first_idx = split_val.find('[')
                    last_idx = split_val.find(']')
                    split_val = split_val[first_idx+1:last_idx]
                    output.append(split_val.replace("'",""))

    else:
        output = (list(curr_hash.keys()))
        
    return output


def get_diff_hash_res(receipe, hash_config) -> List:

    user_settings = dict(get_user_settings(receipe))
    user_resources = get_user_defined_resources(user_settings)
    data = {f"{receipe}": []}
    hash_dict = {}

    existing_hash = _read_json(hash_config)

    for res in user_resources:
        if res in user_settings:
            for i, value in user_settings[res].items():
                print(value)
                hash_val = _hash_fun(value)
                hash_dict[f"{res}.{i}"] = hash_val
    
    data[f"{receipe}"] = [hash_dict]
    print(existing_hash)
    print(data)

    diff_resources = _get_diff_hash(existing_hash, data, receipe)

    return diff_resources


# write configuration to the build hash set
def write_hash_config(hash_dict_set, receipe, hash_file_name):
    output = {}
    user_settings = dict(get_user_settings(receipe))
    user_resources = get_user_defined_resources(user_settings)

    for res in user_resources:
        if res in user_settings:
            for key, val in user_settings[res].items():
                valid_con = f"{res}.{key}"
                if valid_con in hash_dict_set:
                    output[key] = dict(val)
    return {receipe: output}


def run_os_command(command) -> None: 
    try: 
        commands = shlex.split(command)
        if commands[0] == "echo":
            if commands[2] == ">>":
                f = open(commands[-1], mode="a")
            else:
                f = open(command[-1], mode="w")
            code = check_call(commands[:2], stderr=STDOUT, stdout=f)
        else:            
            code = check_call(commands, stderr=STDOUT) 
    except CalledProcessError as e:
        logging.error(str(e))
        code = 1 
    return code 


